"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.OpenAPI = void 0;
const openapi3_ts_1 = require("openapi3-ts");
const zod_1 = require("zod");
const errors_1 = require("./errors");
const file_schema_1 = require("./file-schema");
const helpers_1 = require("./helpers");
const routing_1 = require("./routing");
const mime_1 = require("mime");
const describeSchema = (value, isResponse) => {
    const otherProps = {};
    if (value.isNullable()) {
        otherProps.nullable = true;
    }
    switch (true) {
        case value instanceof zod_1.z.ZodString:
            return {
                ...otherProps,
                ...describeString(value),
            };
        case value instanceof zod_1.z.ZodNumber:
            return {
                ...otherProps,
                ...describeNumber(value)
            };
        case value instanceof zod_1.z.ZodBigInt:
            return { ...otherProps, type: 'integer', format: 'bigint' };
        case value instanceof zod_1.z.ZodBoolean:
            return { ...otherProps, type: 'boolean' };
        case value instanceof zod_1.z.ZodDate:
            return { ...otherProps, type: 'string', format: 'date' };
        case value instanceof zod_1.z.ZodNull:
            // null is not supported https://swagger.io/docs/specification/data-models/data-types/
            return { ...otherProps, type: 'string', nullable: true, format: 'null' };
        case value instanceof zod_1.z.ZodArray:
            return {
                ...otherProps,
                ...describeArray(value._def, isResponse)
            };
        case value instanceof zod_1.z.ZodTuple:
            return {
                ...otherProps,
                ...describeTuple(value, isResponse)
            };
        case value instanceof zod_1.z.ZodRecord:
            return {
                ...otherProps,
                type: 'object',
                additionalProperties: describeSchema(value._def.valueType, isResponse)
            };
        case value instanceof zod_1.z.ZodObject:
            return {
                ...otherProps,
                type: 'object',
                properties: describeObjectProperties(value, isResponse),
                required: Object.keys(value.shape)
                    .filter((key) => !value.shape[key].isOptional())
            };
        case value instanceof zod_1.z.ZodLiteral:
            return {
                ...otherProps,
                type: typeof value._def.value,
                enum: [value._def.value]
            };
        case value instanceof zod_1.z.ZodEnum:
        case value instanceof zod_1.z.ZodNativeEnum:
            return {
                ...otherProps,
                type: typeof Object.values(value._def.values)[0],
                enum: Object.values(value._def.values)
            };
        case value instanceof zod_1.z.ZodTransformer:
        case value instanceof zod_1.z.ZodEffects:
            return {
                ...otherProps,
                ...describeTransformation(value, isResponse)
            };
        case value instanceof zod_1.z.ZodOptional:
        case value instanceof zod_1.z.ZodNullable:
            return {
                ...otherProps,
                ...describeSchema(value.unwrap(), isResponse)
            };
        case value instanceof zod_1.z.ZodIntersection:
            return {
                ...otherProps,
                allOf: [
                    describeSchema(value._def.left, isResponse),
                    describeSchema(value._def.right, isResponse)
                ]
            };
        case value instanceof zod_1.z.ZodUnion:
            return {
                ...otherProps,
                oneOf: value._def.options
                    .map((schema) => describeSchema(schema, isResponse))
            };
        case value instanceof file_schema_1.ZodFile:
            return {
                ...otherProps,
                type: 'string',
                format: value.isBinary ? 'binary' :
                    value.isBase64 ? 'byte' : 'file'
            };
        case value instanceof zod_1.z.ZodAny:
            return {
                ...otherProps,
                format: 'any'
            };
        case value instanceof zod_1.z.ZodUndefined:
        case value instanceof zod_1.z.ZodMap:
        case value instanceof zod_1.z.ZodFunction:
        case value instanceof zod_1.z.ZodLazy:
        case value instanceof zod_1.z.ZodPromise:
        case value instanceof zod_1.z.ZodUnknown:
        case value instanceof zod_1.z.ZodNever:
        case value instanceof zod_1.z.ZodVoid:
        default:
            throw new errors_1.OpenAPIError(`Zod type ${value.constructor.name} is unsupported`);
    }
};
const describeArray = (definition, isResponse) => {
    var _a;
    return ({
        type: 'array',
        items: describeSchema(definition.type, isResponse),
        ...(definition.minLength ? { minItems: definition.minLength.value } : {}),
        ...(definition.maxLength ? { maxItems: (_a = definition.maxLength) === null || _a === void 0 ? void 0 : _a.value } : {})
    });
};
/** @todo improve it when OpenAPI 3.1.0 will be released */
const describeTuple = (schema, isResponse) => {
    const types = schema.items.map((item) => describeSchema(item, isResponse));
    return {
        type: 'array',
        minItems: types.length,
        maxItems: types.length,
        items: {
            oneOf: types,
            format: 'tuple',
            ...(types.length === 0 ? {} : {
                description: types.map((schema, index) => `${index}: ${schema.type}`).join(', ')
            })
        }
    };
};
const describeString = (schema) => {
    const checks = schema._def.checks;
    const isEmail = checks.find(({ kind }) => kind === 'email') !== undefined;
    const isUrl = checks.find(({ kind }) => kind === 'url') !== undefined;
    const isUUID = checks.find(({ kind }) => kind === 'uuid') !== undefined;
    const minLengthCheck = checks.find(({ kind }) => kind === 'min');
    const maxLengthCheck = checks.find(({ kind }) => kind === 'max');
    const regexCheck = checks.find(({ kind }) => kind === 'regex');
    return {
        type: 'string',
        ...(isEmail ? { format: 'email' } : {}),
        ...(isUrl ? { format: 'url' } : {}),
        ...(isUUID ? { format: 'uuid' } : {}),
        ...(minLengthCheck ? { minLength: minLengthCheck.value } : {}),
        ...(maxLengthCheck ? { maxLength: maxLengthCheck.value } : {}),
        ...(regexCheck ? { pattern: `/${regexCheck.regex.source}/${regexCheck.regex.flags}` } : {})
    };
};
const describeNumber = (schema) => {
    const minCheck = schema._def.checks.find(({ kind }) => kind === 'min');
    const isMinInclusive = minCheck ? minCheck.inclusive : true;
    const maxCheck = schema._def.checks.find(({ kind }) => kind === 'max');
    const isMaxInclusive = maxCheck ? maxCheck.inclusive : true;
    return {
        type: schema.isInt ? 'integer' : 'number',
        format: schema.isInt ? 'int64' : 'double',
        minimum: schema.minValue === null ?
            (schema.isInt ? Number.MIN_SAFE_INTEGER : Number.MIN_VALUE) : schema.minValue,
        exclusiveMinimum: !isMinInclusive,
        maximum: schema.maxValue === null ?
            (schema.isInt ? Number.MAX_SAFE_INTEGER : Number.MAX_VALUE) : schema.maxValue,
        exclusiveMaximum: !isMaxInclusive
    };
};
const describeObjectProperties = (schema, isResponse) => {
    return Object.keys(schema.shape).reduce((carry, key) => ({
        ...carry,
        [key]: describeSchema(schema.shape[key], isResponse)
    }), {});
};
const describeTransformation = (value, isResponse) => {
    const input = describeSchema(value._def.schema, isResponse);
    let output = 'undefined';
    if (isResponse && value._def.effects && value._def.effects.length > 0) {
        const effect = value._def.effects.filter((ef) => ef.type === 'transform').slice(-1)[0];
        if (effect && 'transform' in effect) {
            try {
                output = typeof effect.transform(['integer', 'number'].includes(`${input.type}`) ? 0 :
                    'string' === input.type ? '' :
                        'boolean' === input.type ? false :
                            'object' === input.type ? {} :
                                'null' === input.type ? null :
                                    'array' === input.type ? [] : undefined);
            }
            catch (e) { /**/ }
        }
    }
    return {
        ...input,
        ...(['number', 'string', 'boolean', 'null'].includes(output) ? {
            type: output
        } : {})
    };
};
const mimeJson = mime_1.getType('json');
class OpenAPI extends openapi3_ts_1.OpenApiBuilder {
    constructor({ routing, title, version, serverUrl, successfulResponseDescription = 'Successful response', errorResponseDescription = 'Error response' }) {
        super();
        this.addInfo({ title, version }).addServer({ url: serverUrl });
        const cb = (endpoint, fullPath, method) => {
            var _a;
            const operation = {
                responses: {
                    '200': {
                        description: `${method.toUpperCase()} ${fullPath} ${successfulResponseDescription}`,
                        content: endpoint.getPositiveMimeTypes().reduce((carry, mimeType) => ({
                            ...carry,
                            [mimeType]: {
                                schema: describeSchema(endpoint.getPositiveResponseSchema(), true)
                            }
                        }), {}),
                    },
                    '400': {
                        description: `${method.toUpperCase()} ${fullPath} ${errorResponseDescription}`,
                        content: endpoint.getNegativeMimeTypes().reduce((carry, mimeType) => ({
                            ...carry,
                            [mimeType]: {
                                schema: describeSchema(endpoint.getNegativeResponseSchema(), true)
                            }
                        }), {}),
                    }
                }
            };
            if (endpoint.getDescription()) {
                operation.description = endpoint.getDescription();
            }
            if (method === 'get') {
                operation.parameters = [];
                const subject = helpers_1.extractObjectSchema(endpoint.getInputSchema()).shape;
                Object.keys(subject).forEach((name) => {
                    operation.parameters.push({
                        name,
                        in: 'query',
                        required: !subject[name].isOptional(),
                        schema: {
                            ...describeSchema(subject[name], false),
                            description: `${method.toUpperCase()} ${fullPath} parameter`
                        },
                    });
                });
            }
            else {
                operation.requestBody = {
                    content: {
                        [mimeJson]: {
                            schema: {
                                ...describeSchema(endpoint.getInputSchema(), false),
                                description: `${method.toUpperCase()} ${fullPath} request body`
                            }
                        }
                    }
                };
            }
            this.addPath(fullPath, {
                ...(((_a = this.rootDoc.paths) === null || _a === void 0 ? void 0 : _a[fullPath]) || {}),
                [method]: operation,
            });
        };
        routing_1.routingCycle({ routing, cb });
    }
}
exports.OpenAPI = OpenAPI;
//# sourceMappingURL=open-api.js.map