"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.initRouting = exports.routingCycle = exports.DependsOnMethod = void 0;
const endpoint_1 = require("./endpoint");
const errors_1 = require("./errors");
class DependsOnMethod {
    constructor(methods) {
        this.methods = methods;
        Object.keys(methods).forEach((key) => {
            var _a, _b;
            if (key in methods) {
                const endpointMethods = ((_a = methods[key]) === null || _a === void 0 ? void 0 : _a.getMethods()) || [];
                if (!endpointMethods.includes(key)) {
                    throw new errors_1.DependsOnMethodError(`The endpoint assigned to the '${key}' parameter must have at least this method in its specification.\n` +
                        'This error should prevent mistakes during the development process.\n' +
                        'Example:\n\n' +
                        `new ${this.constructor.name}({\n` +
                        `  ${key}: endpointsFactory.build({\n` +
                        `    methods: ['${key}', ` + ((((_b = methods[key]) === null || _b === void 0 ? void 0 : _b.getMethods()) || [])
                        .map((m) => `'${m}'`)
                        .join(', ')
                        || '...') + ']\n' +
                        `    // or method: '${key}'\n` +
                        '    ...\n' +
                        '  })\n' +
                        '});\n');
                }
            }
        });
    }
}
exports.DependsOnMethod = DependsOnMethod;
const routingCycle = ({ routing, cb, parentPath, cors }) => {
    Object.entries(routing).forEach(([path, element]) => {
        path = path.trim();
        if (path.match(/\//)) {
            throw new errors_1.RoutingError('Routing elements should not contain \'/\' character.\n' +
                `The error caused by ${parentPath ? `'${parentPath}' route that has a '${path}'` : `'${path}'`} entry.`);
        }
        const fullPath = `${parentPath || ''}${path ? `/${path}` : ''}`;
        if (element instanceof endpoint_1.AbstractEndpoint) {
            const methods = element.getMethods();
            if (cors) {
                methods.push('options');
            }
            methods.forEach((method) => {
                cb(element, fullPath, method);
            });
        }
        else if (element instanceof DependsOnMethod) {
            Object.entries(element.methods).forEach(([method, endpoint]) => {
                cb(endpoint, fullPath, method);
            });
            if (cors && Object.keys(element.methods).length > 0) {
                const firstEndpoint = Object.values(element.methods)[0];
                cb(firstEndpoint, fullPath, 'options');
            }
        }
        else {
            exports.routingCycle({
                routing: element,
                cb, cors,
                parentPath: fullPath
            });
        }
    });
};
exports.routingCycle = routingCycle;
const initRouting = ({ app, logger, config, routing }) => {
    exports.routingCycle({
        routing,
        cors: config.cors,
        cb: (endpoint, fullPath, method) => {
            app[method](fullPath, async (request, response) => {
                logger.info(`${request.method}: ${fullPath}`);
                await endpoint.execute({ request, response, logger, config });
            });
        }
    });
};
exports.initRouting = initRouting;
//# sourceMappingURL=routing.js.map